# sheldon woodward
# 2/25/18

"""Setuptools File"""

from os.path import abspath, dirname, join
from setuptools import setup


this_dir = abspath(dirname(__file__))
with open(join(this_dir, 'README.rst'), encoding='utf-8') as file:
    long_description = file.read()

install_requires = [r.rstrip('\n') for r in open('requirements.txt').readlines()]


setup(
    name='pymkv',
    description='A Python wrapper for mkvmerge. It provides support for muxing, splitting, linking, chapters, tags, '
                'and attachments through the use of mkvmerge.',
    long_description=long_description,
    url='https://github.com/sheldonkwoodward/pymkv',
    author='Sheldon Woodward',
    author_email='me@sheldonw.com',
    license='MIT',
    packages=['pymkv'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Multimedia :: Video :: Conversion',
        'Topic :: Utilities',
    ],
    keywords='wrapper',
    python_requires='>=3.6',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    install_requires=install_requires,
)
